<?php
/**
 * @author Daniel J. Summers <daniel@bitbadger.solutions>
 * @license MIT
 * @since 1.0.0
 */

declare(strict_types=1);

namespace BitBadger\InspiredByFSharp;

use InvalidArgumentException;

/**
 * The result of an operation; one of two possible outcomes.
 *
 * While exceptions are still possible in F# (and PHP), a more robust model is returning an object that can represent
 * either success or failure. The `Result` type exposes an `OK` state for the successful result of an action and an
 * `Error` state for the unsuccessful result of that action. Neither result value can be `null`; both should be
 * meaningful for the action, and either eventuality should be handled by calling code.
 *
 * The net result _(no pun intended)_ is code that fails gracefully, and calling code that is aware that not every call
 * is successful.
 *
 * @template TOK The type of the OK result
 * @template TError The type of the error result
 */
readonly class Result
{
    /** @var Option<TOK> The OK value for this result */
    private Option $okValue;

    /** @var Option<TError> The error value for this result */
    private Option $errorValue;

    /**
     * Constructor
     *
     * @param TOK|null $okValue The OK value for this result
     * @param TError|null $errorValue The error value for this result
     */
    private function __construct(mixed $okValue = null, mixed $errorValue = null)
    {
        $this->okValue    = Option::of($okValue);
        $this->errorValue = Option::of($errorValue);
    }

    /**
     * Get the value for an `OK` result
     *
     * @return TOK The OK value for this result
     * @throws InvalidArgumentException If the result is an `Error` result
     */
    public function getOK(): mixed
    {
        return $this->okValue->get();
    }

    /**
     * Get the value for an `Error` result
     *
     * @return TError The error value for this result
     * @throws InvalidArgumentException If the result is an `OK` result
     */
    public function getError(): mixed
    {
        return $this->errorValue->get();
    }

    /**
     * Is this result `OK`?
     *
     * @return bool True if the result is `OK`, false if it is `Error`
     */
    public function isOK(): bool
    {
        return $this->okValue->isSome();
    }

    /**
     * Is this result `Error`?
     *
     * @return bool True if the result is `Error`, false if it is `OK`
     */
    public function isError(): bool
    {
        return $this->errorValue->isSome();
    }

    /**
     * Map an `OK` result to another, leaving an `Error` result unmodified
     *
     * @template U The type of the mapping function
     * @param callable(TOK): U $f The mapping function
     * @return Result<U, TError> A transformed `OK` instance or the original `Error` instance
     */
    public function map(callable $f): self
    {
        return $this->isOK() ? self::OK($f($this->getOK())) : $this;
    }

    /**
     * Map an `Error` result to another, leaving an `OK` result unmodified
     *
     * @template U The type of the mapping function
     * @param callable(TError): U $f The mapping function
     * @return Result<TOK, U> A transformed `Error` instance or the original `OK` instance
     */
    public function mapError(callable $f): self
    {
        return $this->isError() ? self::Error($f($this->getError())) : $this;
    }

    /**
     * Execute a function on an `OK` value (if it exists)
     *
     * @param callable(TOK): void $f The function to call
     */
    public function iter(callable $f): void
    {
        if ($this->isOK()) {
            $f($this->getOK());
        }
    }

    /**
     * Transform a `Result`'s `OK` value to an `Option`
     *
     * @return Option<TOK> A `Some` option with the OK value if `OK`, `None` if `Error`
     */
    public function toOption(): Option
    {
        return $this->isOK() ? Option::Some($this->getOK()) : Option::None();
    }

    /**
     * Tap into the `Result` for a secondary action, returning the result
     *
     * @param callable(Result<TOK, TError>): mixed $f The function to run (return value is ignored)
     * @return Result<TOK, TError> The same result provided
     */
    public function tap(callable $f): Result
    {
        $f($this);
        return $this;
    }

    /**
     * Create an `OK` result
     *
     * @param TOK $value The OK value for this result
     * @return Result<TOK, TError> The `OK` result for the value specified
     */
    public static function OK(mixed $value): self
    {
        if (is_null($value)) {
            throw new InvalidArgumentException('Cannot use null as an OK value');
        }
        return new self(okValue: $value);
    }

    /**
     * Create an `Error` result
     *
     * @param TError $value The error value for this result
     * @return Result<TOK, TError> The `Error` result for the value specified
     */
    public static function Error(mixed $value): self
    {
        if (is_null($value)) {
            throw new InvalidArgumentException('Cannot use null as an Error value');
        }
        return new self(errorValue: $value);
    }
}
